@@ -48,11 +48,15 @@
 
 #define ISSET(fd, set)  ((fd) >= 0 && FD_ISSET((fd), (set)))
 #define debug_vector    if (debug_file) print_debug_vector
+#define debug_verbose   if (debug_verbose_flag) debug
+
+#define PROC_MOUNTS_SSHFS   "sshfs#"
 
 extern char **environ;
 
 FILE *debug_file = NULL;
 static char *debug_filename = NULL;
+static int debug_verbose_flag = 0;
 
 static int daemon_pid = -1;
 
@@ -64,6 +68,8 @@ static int mount_expiration = DEFAULT_MOUNT_EXPIRATION;
 static char *mount_cmd = DEFAULT_MOUNT_CMD;
 static char *umount_cmd = DEFAULT_UMOUNT_CMD;
 static char *bind_opt = DEFAULT_BIND_OPT;
+static char *fusermount_cmd  = DEFAULT_FUSERMOUNT_CMD;
+static char *sshfs_cmd =  DEFAULT_SSHFS_CMD;
 
 /** Doubly linked list of mount entries. */
 static struct {
@@ -295,7 +301,7 @@ static void flush(handler_t *data, char *buf, size_t *lenp)
  * Executes argv[0] in a child process.
  * @return -1 on error, 0 otherwise
  */
-static int execute(handler_t *data, char **argv)
+static int execute(handler_t *data, char **argv, int uid, int gid)
 {
 	pid_t pid;
 	int status, err[2];
@@ -318,6 +324,24 @@ static int execute(handler_t *data, char **argv)
 
 		set_debug_name("EXECUTE");
 
+		if (gid >= 0) {
+			debug("Changing to group ID %d", gid);
+
+			if (setgid(gid) < 0) {
+				send_error(data, "Can't change group ID to %d", gid);
+				exit(1);
+			}
+		}
+
+		if (uid >= 0) {
+			debug("Changing to user ID %d", uid);
+
+			if (setuid(uid) < 0) {
+				send_error(data, "Can't change user ID to %d", uid);
+				exit(1);
+			}
+		}
+
 		if (dup2(err[1], STDOUT_FILENO) != STDOUT_FILENO) {
 			send_error(data, "Can't duplicate pipe as stdout");
 			exit(1);
@@ -340,9 +364,13 @@ static int execute(handler_t *data, char **argv)
 
 	while (1) {
 		ssize_t len;
-		char c;
+		char c = '\0';
 
 		len = read(err[0], &c, 1);
+
+		if (c == '\0' && pos == 0)
+			break;
+
 		if (len < 0 && errno == EINTR)
 			continue;
 
@@ -768,6 +796,8 @@ static void mnt_list_del(mount_t *mnt)
 static int do_mount(handler_t *data, const mount_info_t *mi)
 {
 	char *argv[8];
+	int uid = -1;
+	int gid = -1;
 
 	memset(argv, 0, sizeof (argv));
 
@@ -798,21 +828,48 @@ static int do_mount(handler_t *data, const mount_info_t *mi)
 
 		break;
 
+	case MTYPE_SSH:
+		argv[0] = sshfs_cmd;
+
+		argv[3] = mi->opts ? "-o" : NULL;
+		argv[4] = mi->opts;
+
+		if (mi->opts) {
+			debug("Executing: %s %s %s -o %s", argv[0], argv[1], argv[2], argv[4]);
+		} else {
+			debug("Executing: %s %s %s", argv[0], argv[1], argv[2]);
+		}
+
+		if (!data->param.ids || data->param.ids->len < 2) {
+			errno = 0;
+			send_error(data, "Valid IDS parameter required");
+			return -1;
+		}
+
+		uid = data->param.ids->vec[0];
+		gid = data->param.ids->vec[1];
+		break;
+
 	default:
 		errno = EINVAL;
 		return -1;
 	}
 
-	return execute(data, argv);
+	return execute(data, argv, uid, gid);
 }
 
-static int do_unmount(handler_t *data, char *point)
+static int do_unmount(handler_t *data, const mount_info_t *mi)
 {
-	char *argv[] = { umount_cmd, point, NULL };
+	char *argv[] = { umount_cmd, mi->point, NULL, NULL };
 
-	debug("Executing: %s %s", argv[0], argv[1]);
-
-	return execute(data, argv);
+	/* Handle SSHFS as special case, needs to be unmounted using fusermount */
+	if (mi->type == MTYPE_SSH) {
+		argv[0] = fusermount_cmd;
+		argv[1] = "-u";
+		argv[2] = mi->point;
+	}
+	debug("Executing: %s %s %s", argv[0], argv[1], argv[2]);
+	return execute(data, argv, -1, -1);
 }
 
 static mount_t *mnt_create(handler_t *data, mount_info_t *mi)
@@ -838,8 +895,9 @@ static mount_t *mnt_create(handler_t *data, mount_info_t *mi)
 }
 
 /**
- * Checks /proc/mounts if mount_info_t is already mounted.
+ * Checks /proc/mounts if mount point is already mounted.
  * @param data handler state
+ * @param point mount point to compare against
  * @return -1 on error, 0 otherwise
  */
 static int is_mounted(handler_t *data, const char *point)
@@ -861,7 +919,7 @@ static int is_mounted(handler_t *data, const char *point)
 	while (1) {
 		char *device, *point2tmp, *point2;
 
-		if (read_line(file, buf, sizeof (buf)) < 0)
+		if (read_line_nocomments(file, buf, sizeof (buf)) < 0)
 			break;
 
 		split_string(buf, &device, &point2tmp, NULL);
@@ -941,7 +999,7 @@ static int remove_mount(handler_t *data, const mount_info_t *mi)
 			send_error(data, "Warning: %s usage is %d", mnt->info.point, mnt->usage);
 	}
 
-	if (mounted && do_unmount(data, mi->point) < 0)
+	if (mounted && do_unmount(data, mi) < 0)
 		return -1;
 
 	return 0;
@@ -956,8 +1014,13 @@ static void release_mount(mount_t *mnt)
 
 	debug("Releasing mount %s (%d users remain)", mnt->info.point, mnt->usage);
 
-	if (mnt->usage <= 0 && mount_expiration >= 0)
-		mnt->expiration = time(NULL) + mount_expiration;
+	if (mnt->usage <= 0 && mount_expiration >= 0) {
+		/* Force bind mounts to expire immediately */
+		if (mnt->info.type == MTYPE_BIND)
+			mnt->expiration = 0;
+		else
+			mnt->expiration = time(NULL) + mount_expiration;
+	}
 }
 
 /**
@@ -990,7 +1053,8 @@ static void expire_mounts(void)
 			else
 				all_mounts.head = curr->next;
 
-			do_unmount(NULL, curr->info.point);
+			// TODO Check if unmount was succesful
+			do_unmount(NULL, &curr->info);
 			mnt_free(curr);
 		}
 
@@ -998,6 +1062,137 @@ static void expire_mounts(void)
 	}
 }
 
+/**
+ * Parse remote filesystem path from /proc/mounts for mount point
+ * @param data handler state
+ * @param point mount point to compare against
+ * @param cpy remote filesystem path
+ * @return string length on success, -1 otherwise
+ */
+static int get_remote(handler_t *data, const char *point, char *cpy)
+{
+	FILE *file;
+	char buf[1024], buf1[PATH_MAX], buf2[PATH_MAX], *point1;
+
+	point1 = realpath(point, buf1);
+	if (!point1)
+		return -1;
+
+	file = fopen(MOUNTS_FILE, "r");
+	if (!file) {
+		send_error(data, "Can't open " MOUNTS_FILE);
+		return -1;
+	}
+
+	while (1) {
+		char *device, *point2tmp, *point2;
+
+		if (read_line_nocomments(file, buf, sizeof (buf)) < 0)
+			break;
+
+		split_string(buf, &device, &point2tmp, NULL);
+		if (!device || !point2tmp)
+			continue;
+
+		point2 = realpath(point2tmp, buf2);
+		if (!point2) {
+			error("Can't get real path of %s", point2tmp);
+			continue;
+		}
+
+		if (strcmp(point1, point2) != 0)
+			continue;
+
+		mount_t *mnt = mnt_list_find(point);
+
+		/* Sample sshfs entry:
+		 * sshfs#you@host:/path / fuse nosuid,nodev,max_read=65536 0 0
+		 */
+		if (mnt && (mnt->info.type == MTYPE_SSH &&
+		            strncmp(device, PROC_MOUNTS_SSHFS, strlen(PROC_MOUNTS_SSHFS)) == 0)) {
+			strcat(cpy, device + strlen(PROC_MOUNTS_SSHFS));
+			debug("Remote sshfs mount point is '%s'", cpy);
+			fclose(file);
+			return strlen(cpy);
+		}
+
+		/* Sample nfs entry:
+		 * host:/path / nfs rw,addr=host 0 0
+		 */
+		if (mnt && (mnt->info.type == MTYPE_NFS)) {
+			strcat(cpy, device);
+			debug("Remote nfs mount point is '%s'", cpy);
+			fclose(file);
+			return strlen(cpy);
+		}
+	}
+
+	fclose(file);
+	return -1;
+}
+
+/**
+ * Expire mount points if remote mount point has changed.
+ * @param data handler state
+ * @param mi mount point to compare against
+ * @return -1 on error, 0 otherwise
+ */
+static int expire_changed(handler_t *data, const mount_info_t *mi)
+{
+	mount_t *mnt;
+	char remote_path[PATH_MAX] = { 0 };  /* On Linux path max is 4096 */
+	int len;
+
+	mnt = mnt_list_find(mi->point);
+	if (mnt == NULL)
+		return 0;
+
+	/* The remote path for a bind mount can be aliased to the original
+	 * device for mount points (not the directory originally bound), so
+	 * get_remote() will fail for those; and, we don't even check binds
+	 * below, so bail early.
+	 */
+	if (mi->type == MTYPE_BIND)
+		return 0;
+
+	len = get_remote(data, mi->point, remote_path);
+	if (len <= 0)
+		return len;
+
+	/* TODO: Add check also for NFS mounts */
+	if ((mi->type == MTYPE_NFS || mi->type == MTYPE_SSH) &&
+	    strcmp(remote_path, mi->device) != 0) {
+		mount_t *curr;
+		int exprcnt = 0;
+
+		warn("Remote mount point for '%s' has changed", mi->point);
+
+		/* Match target root and mount points under target root.
+		 * Tries to set expires also for bind mounts under the matching
+		 * mount point.
+		 */
+		for (curr = all_mounts.tail; curr != NULL; curr = curr->prev) {
+			if (strncmp(mi->point, curr->info.point, strlen(mi->point) - 1) != 0)
+				continue;
+
+			if (curr->usage <= 0) {
+				debug("Setting %s as expired", curr->info.point);
+				curr->expiration = 0;
+				++exprcnt;
+			} else {
+				send_error(data, "Warning: %s remote mount point has changed, has usage %d",
+				           mnt->info.point, mnt->usage);
+				return -1;
+			}
+		}
+
+		if (exprcnt > 0)
+			expire_mounts();
+	}
+
+	return 0;
+}
+
 /**
  * Unmounts all entries in the MOUNTS list.
  */
@@ -1011,7 +1206,8 @@ static void unmount_all(void)
 
 		debug("Unmounting %s", mnt->info.point);
 
-		do_unmount(NULL, mnt->info.point);
+		// TODO Check if unmount was succesful
+		do_unmount(NULL, &mnt->info);
 		mnt_free(mnt);
 
 		mnt = prev;
@@ -1096,7 +1292,7 @@ static void send_data(handler_t *data, output_desc_t *out)
 	}
 
 	if (len) {
-		debug("Sending %s DATA packet (%d bytes)",
+		debug_verbose("Sending %s DATA packet (%d bytes)",
 		      output_desc_is_stdout(out) ? "OUT" : "ERR", len);
 
 		if (write_buf_packet(data->sd, out->data_type, len, data->tmp_buf) < 0) {
@@ -1127,7 +1323,7 @@ static void send_request(handler_t *data, input_desc_t *in)
 	if (in->wait)
 		return;
 
-	debug("Sending IN REQ packet");
+	debug_verbose("Sending IN REQ packet");
 
 	if (write_enum(data->sd, in->req_type) < 0) {
 		error("Can't write packet to socket");
@@ -1146,7 +1342,7 @@ _error:
  */
 static void write_buffer(handler_t *data, input_desc_t *in)
 {
-	debug("Writing buffer to stdin");
+	debug_verbose("Writing buffer to stdin");
 
 	if (buf_write_out(in->buf, &in->fd) < 0)
 		debug("Stdin is closed (%s)", strerror(errno));
@@ -1169,7 +1365,7 @@ static void receive_stream(handler_t *data)
 	}
 
 	if (len > 0) {
-		debug("Receiving IN DATA packet (%d bytes)", len);
+		debug_verbose("Receiving IN DATA packet (%d bytes)", len);
 		if (buf_read_in(data->in.buf, data->sd, len) < 0) {
 			send_error(data, "Can't read IN DATA packet to buffer");
 			goto _error;
@@ -1207,12 +1403,12 @@ static int receive_packet(handler_t *data)
 		return 0;
 
 	case PTYPE_OUT_REQ:
-		debug("Receiving OUT REQ packet");
+		debug_verbose("Receiving OUT REQ packet");
 		data->out.req = BUFFER_SIZE;
 		return 0;
 
 	case PTYPE_ERR_REQ:
-		debug("Receiving ERR REQ packet");
+		debug_verbose("Receiving ERR REQ packet");
 		data->err.req = BUFFER_SIZE;
 		return 0;
 
@@ -1254,7 +1450,7 @@ static void print_status(int pid, int status)
 #define debug_bool(m, b)        if (debug_file) print_debug_bool(m, b)
 
 #define debug_status() \
-	if (debug_file) { \
+	if (debug_verbose_flag && debug_file) { \
 		print_debug_bool("Process alive  = %s", alive); \
 		print_debug_bool("Stdin open     = %s", data->in.fd >= 0); \
 		print_debug_bool("Stdin waiting  = %s", data->in.wait); \
@@ -1269,7 +1465,7 @@ static void print_status(int pid, int status)
 	}
 
 #define debug_select(msg) \
-	if (debug_file) { \
+	if (debug_verbose_flag && debug_file) { \
 		print_debug(msg " [%3s %2s %3s %3s][%2s]", \
 			    ISSET(data->sd,      &readfds) ? "NET" : "", \
 			    ISSET(data->in.fd,   &readfds) ? "IN"  : "", \
@@ -1692,11 +1888,18 @@ static void execute_command(handler_t *data)
 				}
 
 		gcount = data->param.ids->len - 2;
-		if (gcount > NGROUPS_MAX)
-			gcount = NGROUPS_MAX;
 
-		for (i = 0; i < gcount; ++i)
-			groups[i] = data->param.ids->vec[i + 2];
+		/*
+		 * If no groups are passed in, then use the local
+		 * group mapping.
+		 */
+		if (gcount > 0) {
+			if (gcount > NGROUPS_MAX)
+				gcount = NGROUPS_MAX;
+
+			for (i = 0; i < gcount; ++i)
+				groups[i] = data->param.ids->vec[i + 2];
+		}
 
 		debug("Changing gid to %d", gid);
 		if (setgid(gid) < 0) {
@@ -1704,12 +1907,16 @@ static void execute_command(handler_t *data)
 			exit(1);
 		}
 
-		debug("Setting %d supplementary gids", gcount);
-		if (setgroups(gcount, groups) < 0) {
-			send_error(data, "Can't set supplementary group IDs");
+		if (gcount > 0) {
+			debug("Setting %d supplementary gids", gcount);
+			if (setgroups(gcount, groups) < 0) {
+				send_error(data, "Can't set supplementary group IDs");
 #ifndef DEBUG
-			exit(1);
+				exit(1);
 #endif
+			}
+		} else {
+			debug("Using groups of user");
 		}
 
 		debug("Changing uid to %d", uid);
@@ -2059,6 +2266,14 @@ static int handler_startup(handler_t *data, mount_t ***mounts_ptr)
 			return -1;
 		}
 
+		for (i = 0; i < mount_count; ++i)
+			if (expire_changed(data, data->param.mounts[i]) < 0) {
+				errno = 0;
+				send_error(data, "Can't expire remote mount point: %s",
+				           data->param.mounts[i]->point);
+				return -1;
+			}
+
 		for (i = 0; i < mount_count; ++i) {
 			mounts[i] = add_mount(data, data->param.mounts[i]);
 			if (!mounts[i]) {
@@ -2463,6 +2678,8 @@ static void usage(char *progname)
 				 " [-e|--mount-expiration <minutes>|none]"
 				 " [-m|--mount-bin <path>]"
 				 " [-u|--umount-bin <path>]"
+				 " [-S|--sshfs-bin <path>]"
+				 " [-F|--fusermount-bin <path>]"
 				 " [-t|--mount-tab <path>]"
 				 " [-b|--bind-opt <options>]\n"
 			"       %s add <address>\n"
@@ -2478,7 +2695,7 @@ static void usage(char *progname)
  */
 static void read_args(char *progname, int argc, char **argv)
 {
-	const char *const optstring = "hvp:lnird:e:m:u:b:";
+	const char *const optstring = "hvp:lnird:Ve:m:u:b:";
 	struct option longopts[] = {
 		{ "help",             no_argument,       0, 'h' },
 		{ "version",          no_argument,       0, 'v' },
@@ -2487,9 +2704,12 @@ static void read_args(char *progname, int argc, char **argv)
 		{ "no-sandbox",       no_argument,       0, 'n' },
 		{ "allow-root",       no_argument,       0, 'r' },
 		{ "debug",            required_argument, 0, 'd' },
+		{ "debug-verbose",    no_argument,       0, 'V' },
 		{ "mount-expiration", required_argument, 0, 'e' },
 		{ "mount-bin",        required_argument, 0, 'm' },
 		{ "umount-bin",       required_argument, 0, 'u' },
+		{ "sshfs-bin",        required_argument, 0, 'S' },
+		{ "fusermount-bin",   required_argument, 0, 'F' },
 		{ "bind-opt",         required_argument, 0, 'b' },
 		{ 0 }
 	};
@@ -2524,6 +2744,10 @@ static void read_args(char *progname, int argc, char **argv)
 			debugname = optarg;
 			break;
 
+		case 'V':
+			debug_verbose_flag = TRUE;
+			break;
+
 		case 'e':
 			exp_str = optarg;
 			break;
@@ -2536,6 +2760,14 @@ static void read_args(char *progname, int argc, char **argv)
 			umount_cmd = optarg;
 			break;
 
+		case 'S':
+			sshfs_cmd = optarg;
+			break;
+
+		case 'F':
+			fusermount_cmd = optarg;
+			break;
+
 		case 'b':
 			bind_opt = optarg;
 			default_bind_opt = FALSE;
